import asyncio
import datetime
import enum
import json
import logging
from typing import Optional, Type, Generic, List, Tuple, Any, Dict

import aiomysql  # type: ignore
import pymysql

from ..base import T
from ..conversion import SHCJsonEncoder, from_json
from .generic import PersistenceVariable
from ..supervisor import AbstractInterface, InterfaceStatus, ServiceStatus

logger = logging.getLogger(__name__)


class MySQLPersistence(AbstractInterface):
    def __init__(self, **kwargs):
        super().__init__()
        # see https://aiomysql.readthedocs.io/en/latest/connection.html#connection for valid parameters
        self.connect_args = kwargs
        self.pool: Optional[aiomysql.Pool] = None
        self.pool_ready = asyncio.Event()
        self.variables: Dict[str, MySQLPersistenceVariable] = {}

    async def start(self) -> None:
        logger.info("Creating MySQL connection pool ...")
        self.pool = await aiomysql.create_pool(**self.connect_args)
        self.pool_ready.set()

    async def stop(self) -> None:
        if self.pool is not None:
            logger.info("Closing all MySQL connections ...")
            self.pool.close()
            await self.pool.wait_closed()

    async def get_status(self) -> "InterfaceStatus":
        if not self.pool_ready.is_set():
            return InterfaceStatus(ServiceStatus.CRITICAL, "Interface not started yet")
        assert(isinstance(self.pool, aiomysql.Pool))
        free_connections = self.pool.freesize
        try:
            async with self.pool.acquire() as conn:
                async with conn.cursor() as cur:
                    await cur.execute("SELECT * from `log` WHERE FALSE")
        except pymysql.err.MySQLError as e:
            return InterfaceStatus(ServiceStatus.CRITICAL, str(e), {'free_connections': free_connections})
        return InterfaceStatus(ServiceStatus.OK, "", {'free_connections': free_connections})

    def variable(self, type_: Type, name: str, log: bool = True) -> "MySQLPersistenceVariable":
        if name in self.variables:
            variable = self.variables[name]
            if variable.type is not type_:
                raise ValueError("MySQL persistence variable with name {} has already been defined with type {}"
                                 .format(name, variable.type.__name__))
            return variable
        return MySQLPersistenceVariable(self, type_, name, log)

    def __repr__(self) -> str:
        return "{}({})".format(self.__class__.__name__, {k: v for k, v in self.connect_args.items()
                                                         if k in ('host', 'db', 'port', 'unix_socket')})


class MySQLPersistenceVariable(PersistenceVariable, Generic[T]):
    def __init__(self, interface: MySQLPersistence, type_: Type[T], name: str, log: bool):
        super().__init__(type_, log)
        self.interface = interface
        self.name = name
        self.log = log

    async def _write_to_log(self, value: T):
        column_name = self._type_to_column(type(value))
        value = self._into_mysql_type(value)
        await self.interface.pool_ready.wait()
        assert(self.interface.pool is not None)
        async with self.interface.pool.acquire() as conn:
            async with conn.cursor() as cur:
                if self.log:
                    await cur.execute("INSERT INTO `log` (`name`, `ts`, `{}`) VALUES (%s, %s, %s)".format(column_name),
                                      (self.name, datetime.datetime.now(datetime.timezone.utc), value))
                else:
                    await cur.execute("UPDATE `log` SET `ts` = %s, `{}` = %s WHERE `name` = %s".format(column_name),
                                      (datetime.datetime.now(datetime.timezone.utc), value, self.name))
            await conn.commit()

    async def _read_from_log(self) -> Optional[T]:
        column_name = self._type_to_column(self.type)
        await self.interface.pool_ready.wait()
        assert(self.interface.pool is not None)
        async with self.interface.pool.acquire() as conn:
            async with conn.cursor() as cur:
                await cur.execute("SELECT `{}` from `log` WHERE `name` = %s ORDER BY `ts` DESC LIMIT 1"
                                  .format(column_name),
                                  (self.name,))
                value = await cur.fetchone()
        if value is None:
            return None
        logger.debug("Retrieved value %s for %s from %s", value, self, self.interface)
        return self._from_mysql_type(value[0])

    async def retrieve_log(self, start_time: datetime.datetime, end_time: datetime.datetime,
                           include_previous: bool = True) -> List[Tuple[datetime.datetime, T]]:
        column_name = self._type_to_column(self.type)
        await self.interface.pool_ready.wait()
        assert(self.interface.pool is not None)
        async with self.interface.pool.acquire() as conn:
            async with conn.cursor() as cur:
                if include_previous:
                    await cur.execute(
                        "(SELECT `ts`, `{0}` from `log` WHERE `name` = %s AND `ts` <= %s "
                        " ORDER BY `ts` DESC LIMIT 1) "
                        "UNION (SELECT `ts`, `{0}` from `log` WHERE `name` = %s AND `ts` > %s AND `ts` < %s "
                        " ORDER BY `ts` ASC)".format(column_name),
                        (self.name, start_time, self.name, start_time, end_time))
                else:
                    await cur.execute(
                        "SELECT `ts`, `{0}` from `log` WHERE `name` = %s AND `ts` >= %s AND `ts` < %s "
                        "ORDER BY `ts` ASC".format(column_name),
                        (self.name, start_time, end_time))

                return [(row[0].replace(tzinfo=datetime.timezone.utc), self._from_mysql_type(row[1]))
                        for row in await cur.fetchall()]

    @classmethod
    def _type_to_column(cls, type_: Type) -> str:
        if issubclass(type_, (int, bool)):
            return 'value_int'
        elif issubclass(type_, float):
            return 'value_float'
        elif issubclass(type_, str):
            return 'value_str'
        elif issubclass(type_, enum.Enum):
            return cls._type_to_column(type(next(iter(type_.__members__.values())).value))
        else:
            return 'value_str'

    @classmethod
    def _into_mysql_type(cls, value: Any) -> Any:
        if isinstance(value, bool):
            return value
        elif isinstance(value, int):
            return int(value)
        elif isinstance(value, float):
            return float(value)
        elif isinstance(value, str):
            return str(value)
        elif isinstance(value, enum.Enum):
            return cls._into_mysql_type(value.value)
        else:
            return json.dumps(value, cls=SHCJsonEncoder)

    def _from_mysql_type(self, value: Any) -> Any:
        if issubclass(self.type, (bool, int, float, str, enum.Enum)):
            return self.type(value)
        else:
            return from_json(self.type, json.loads(value))

    def __repr__(self):
        return "<{} '{}'>".format(self.__class__.__name__, self.name)



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
import asyncio
import enum
import functools
import logging
import signal
from typing import Set, NamedTuple, Dict, Any, Union, Iterable

from .timer import timer_supervisor
from .variables import read_initialize_variables

logger = logging.getLogger(__name__)

_REGISTERED_INTERFACES: Set["AbstractInterface"] = set()

event_loop = asyncio.get_event_loop()
_EXIT_CODE = 0
_SHC_STOPPED = asyncio.Event(loop=event_loop)
_SHC_STOPPED.set()


class ServiceCriticality(enum.Enum):
    """
    Enum of possible criticality values of interfaces.
    """
    INFO = 0
    WARNING = 1
    CRITICAL = 2


class AbstractInterface(metaclass=abc.ABCMeta):
    """
    Abstract base class for all SHC interface implementations

    An interface is an object that is 'started' at SHC startup. This is typically used to run interfaces to the outside
    world, using network connections (or serial device connections) and background asyncio Tasks for handling incoming
    messages and forwarding them to *Subscribable* objects.

    If an interface inherits from this base class, it is automatically registered for startup via :func:`main`.

    :ivar criticality: Defines to which extend the interface's status is considered when determining the overall SHC
        system state, e.g. when reporting to a monitoring system or creating alerts in a user interface. A critical
        failure of a *CRITICAL* system is considered a critical state, whereas a critical failure of an *INFO* system
        only triggers an information message.
    """
    criticality: ServiceCriticality = ServiceCriticality.CRITICAL

    def __init__(self):
        register_interface(self)

    @abc.abstractmethod
    async def start(self) -> None:
        """
        This coroutine is called once on each interface when the SHC application is started via :func:`main`.

        It may be used to initialize network connections and start background Tasks for message handling and event
        generation. The `start` coroutines of all interfaces are started in parallel. Timers and Variable's are only
        initialized when *all* `start` coroutines have returned successfully.

        The method should await the completion of the interface startup, i.e. only return when the interface is fully
        functional. In case of an error, this method may raise and exception or call :func:`interface_failure` to
        terminate the SHC application.
        """
        pass

    @abc.abstractmethod
    async def stop(self) -> None:
        """
        This coroutine is called once on each interface to shut down the SHC application.

        This may happen when a SIGTERM (or similar) is received or when a critical error in an interface is reported
        (via :func:`interface_failure` or an exception raised by an interface's :meth:`start` method). Thus, the `stop`
        method may even be called while :meth:`start` is still running. It shall be able to handle that situation and
        stop the half-started interface.

        In any case, the `stop` method shall cause the termination of all pending background task of the interface
        within a few seconds and only return when all tasks have terminated.
        """
        pass

    async def get_status(self) -> "InterfaceStatus":
        """
        Get the current status of the interface for monitoring purposes.

        This is especially required for interfaces that do not shut down the SHC application via
        :func:`interface_failure` on every disruption, but instead keep trying to recover operation. In the meantime,
        ServiceStatus.CRITICAL shall be reported via this method.
        """
        return InterfaceStatus()


class ServiceStatus(enum.Enum):
    """
    Enum of possible service status, derived from Nagios/Icinga status.
    """
    OK = 0
    WARNING = 1
    CRITICAL = 2
    UNKNOWN = 3


class InterfaceStatus(NamedTuple):
    """
    Interface status information as returned by :meth:`AbstractInterface.get_status`.

    Contains the overall interface status (:attr:`status`), a human readable :attr:`message`, typically describing the
    error if any, and a map of :attr:`indicators`, which contain interface-specific performance values.
    """
    status: ServiceStatus = ServiceStatus.OK  #: Overall status of the interface.
    message: str = ""  #: A textual description of the error. E.g. an error message, if status != ServiceStatus.OK
    #: Additional monitoring indicators like performance values, identified by a unique string.
    indicators: Dict[str, Union[bool, int, float, str]] = {}


def register_interface(interface: AbstractInterface) -> None:
    _REGISTERED_INTERFACES.add(interface)


def get_interfaces() -> Iterable[AbstractInterface]:
    return _REGISTERED_INTERFACES


async def interface_failure(interface_name: str = "n/a") -> None:
    """
    Shut down the SHC application due to a critical error in an interface

    This coroutine shall be called from an interface's background Task on a critical failure.
    It will shut down the SHC system gracefully and lets the Python process return with exit code 1.

    :param interface_name: String identifying the interface which caused the shutdown.
    """
    logger.warning("Shutting down SHC due to error in interface %s", interface_name)
    global _EXIT_CODE
    _EXIT_CODE = 1
    asyncio.create_task(stop())


async def run():
    _SHC_STOPPED.clear()
    logger.info("Starting up interfaces ...")
    # TODO catch errors and stop() in case of an exception
    await asyncio.gather(*(interface.start() for interface in _REGISTERED_INTERFACES))
    logger.info("All interfaces started successfully. Initializing variables ...")
    await read_initialize_variables()
    logger.info("Variables initialized successfully. Starting timers ...")
    await timer_supervisor.start()
    logger.info("Timers initialized successfully. SHC startup finished.")
    # Now, keep this task awaiting until SHC is stopped via stop()
    await _SHC_STOPPED.wait()


async def stop():
    logger.info("Shutting down interfaces ...")
    await asyncio.gather(*(interface.stop() for interface in _REGISTERED_INTERFACES), timer_supervisor.stop(),
                         return_exceptions=True)
    _SHC_STOPPED.set()


def handle_signal(sig: int, loop: asyncio.AbstractEventLoop):
    logger.info("Got signal {}. Initiating shutdown ...".format(sig))
    loop.create_task(stop())


def main() -> int:
    """
    Main entry point for running an SHC application

    This function starts an asyncio event loop to run the timers and interfaces. It registers signal handlers for
    SIGINT, SIGTERM, and SIGHUP to shut down all interfaces gracefully when such a signal is received. The `main`
    function blocks until shutdown is completed and returns the exit code. Thus, it should be used with
    :func:`sys.exit`::

        import sys
        import shc

        # setup interfaces, connect objects, etc.

        sys.exit(shc.main())

    A shutdown can also be triggered by a critical error in an interface (indicated via :func:`interface_failure`), in
    which case the the exit code will be != 0.

    :return: application exit code to be passed to sys.exit()
    """
    for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGHUP):
        event_loop.add_signal_handler(sig, functools.partial(handle_signal, sig, event_loop))
    event_loop.run_until_complete(run())
    return _EXIT_CODE



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
import asyncio
import logging
import threading
from typing import List, Any, Tuple, Dict, Optional, Union, Iterable

import mido  # type: ignore

from ..base import Subscribable, Writable, T
from ..datatypes import RangeUInt8
from ..supervisor import stop, AbstractInterface

logger = logging.getLogger(__name__)


class MidiInterface(AbstractInterface):
    """
    An SHC interface for connecting with MIDI devices.

    This interface is primarily designed to control SHC applications with MIDI DAW controllers such as the *Behringer
    X-Touch*, *Icon QCon*, *DJ Techtools Midi Fighter* or similar, which provide (motorized) faders, rotary encoders
    and/or push buttons. These interfaces typically send MIDI Note on/off and control change messages when the user
    uses the controls. In return, the controller's hardware feedback (motor faders, LEDs) can be updated by sending the
    corresponding MIDI events to the controller.

    Thus, the `MidiInterface` allows to create the following types of *Connectable* objects:

    * :meth:`note_on_off`: :class:`bool`-type for sending/receiving Note on/off events for a specific note number (for
      push buttons)
    * :meth:`note_velocity`: range-type (:class:`RangeUint8 <shc.datatypes.RangeUInt8>`) for sending/receiving Note
      on/off events with a value encoded in the velocity parameter of a specific note number
    * :meth:`control_change`: range-type (:class:`RangeUint8 <shc.datatypes.RangeUInt8>`) for sending/receiving Control
      Change events with a for a specific control.

    However, the Interface features might also be suited to control any MIDI gear, which reacts to these MIDI events.

    The outbound MIDI channel number can be set globally for all *Connectables* of the interface (``send_channel``).
    Inbound MIDI events can be filtered for a specific MIDI channel or a list of channels (``receive_channel``).

    The interface can be used in a bidirectional mode, input-only mode (no output port; values `written` to the
    *Connectable* objects are dropped) or output-only mode (no input port).

    :param input_port_name: Name of the inbound MIDI port. If not specified, the interface works in output-only mode.
        Use ``mido.get_input_names()`` to get a list of output port names.
    :param output_port_name: Name of the outbound MIDI port. If not specified, the interface works in input-only mode.
        Use ``mido.get_output_names()`` to get a list of output port names.
    :param send_channel: MIDI channel number [0-15] for all outbound MIDI messages. Defaults to 0.
    :param receive_channel: If specified, it allows to filter the incomming MIDI events by MIDI channel. Either a single
        channel number [0-15] or a list of channel numbers which should be processed. Other than the specified MIDI
        channel(s) are ignored.
    """
    def __init__(self, input_port_name: Optional[str] = None,
                 output_port_name: Optional[str] = None,
                 send_channel: int = 0,
                 receive_channel: Union[None, int, Iterable[int]] = None) -> None:
        if not input_port_name and not output_port_name:
            raise ValueError("Either MIDI input port name or output port name must be specified.")
        super().__init__()

        self.input_port_name = input_port_name
        self.output_port_name = output_port_name
        self.send_channel = send_channel
        self.receive_channel = receive_channel

        self._output_queue: asyncio.Queue[Optional[mido.Message]]
        self._input_queue: asyncio.Queue[mido.Message]

        self._variable_map: Dict[Tuple[str, int], AbstractMidiVariable] = {}

    async def start(self) -> None:
        loop = asyncio.get_event_loop()
        # We must initialize the Queues and Events here to make sure that they are assigned to the correct event loop
        self._output_queue = asyncio.Queue()
        self._input_queue = asyncio.Queue()
        self._send_thread_stopped = asyncio.Event()
        self._send_thread_stopped.set()

        if self.output_port_name:
            send_thread = threading.Thread(target=self._send_thread, name="shc.midi.send_thread", args=(loop,))
            send_thread.start()
            self._send_thread_stopped.clear()

        def incoming_message_callback(message):
            loop.call_soon_threadsafe(self._dispatch_message, message)

        if self.input_port_name:
            self.input_port = mido.open_input(self.input_port_name, callback=incoming_message_callback)

    async def stop(self) -> None:
        logger.info('Stopping MIDI interface.')

        if self.input_port_name:
            logger.debug('First, closing down mido input_port ...')
            await asyncio.get_event_loop().run_in_executor(None, self.input_port.close)

        if self.output_port_name:
            logger.debug('Sending None value to _send_thread() to shut it down ...')
            await self._output_queue.put(None)
            await self._send_thread_stopped.wait()
        logger.debug('MIDI interface shutdown finished.')

    def _dispatch_message(self, message) -> None:
        logger.debug('Received MIDI message: %s', message)
        msg_type = message.type
        channel = message.channel
        # Filter by MIDI event type
        if msg_type not in ('note_on', 'note_off', 'control_change'):
            logger.debug('Unsupported MIDI message type %s', msg_type)
            return
        # Filter by MIDI channel
        if isinstance(self.receive_channel, int):
            if channel != self.receive_channel:
                logger.debug('Ignoring message from wrong input channel %s', channel)
                return
        elif self.receive_channel is not None and channel not in self.receive_channel:
            logger.debug('Ignoring message from wrong input channel %s', channel)
            return

        # Check if a *Connectable* object is registered for this type/index combination
        key = message.control if msg_type == 'control_change' else message.note
        variable = self._variable_map.get((msg_type, key))
        if variable is None:
            logger.debug('MIDI message is not assigned to any variable')
            return

        # Dispatch new value in a new asyncio Task
        try:
            variable._incoming_message(message)
        except Exception as e:
            logger.error("Error while dispatching incoming MIDI message %s", message, exc_info=e)

    def _send_thread(self, loop: asyncio.AbstractEventLoop) -> None:
        """
        Entry point for the daemon thread for sending outgoing MIDI messages to the MIDI port.
        """
        try:
            logger.debug("Starting _send_thread(). Opening output_port for %s", self.output_port_name)
            output_port = mido.open_output(self.output_port_name)
            logger.debug("Output port to %s opened.", self.output_port_name)
            while True:
                message: mido.Message = asyncio.run_coroutine_threadsafe(self._output_queue.get(), loop).result()
                if message is None:
                    logger.debug("_send_thread got None value. Exiting send loop.")
                    break
                logger.debug('Sending MIDI message: %s', message)
                output_port.send(message)

            logger.info('Closing mido output_port ...')
            output_port.close()
            logger.debug('_send_thread() done.')
        except Exception as e:
            logger.critical("Exception in MIDI send thread. Shutting down SHC ...", exc_info=e)
            asyncio.run_coroutine_threadsafe(stop(), loop)
        finally:
            loop.call_soon_threadsafe(self._send_thread_stopped.set)

    def _send_message(self, message: mido.Message) -> None:
        """
        Internal method used by the Variable objects to send a MIDI message to the output port via the output queue.

        This method takes care of dropping the messages if this interface works in input-only mode.
        :param message: The MIDI message to send to the output port
        """
        if self.output_port_name:
            self._output_queue.put_nowait(message)

    def note_on_off(self, note: int, emulate_toggle: bool = False,
                    off_output_velocity: int = 0, on_output_velocity: int = 127) -> "NoteOnOffVariable":
        """
        Create a *Connectable* object for this MIDI interface which handles Note on/off events as boolean values.

        The returned object is *Writable* and *Subscribable*. It will publish a `True` value when a Note on event for
        the specified Note is received and a `False` for each Note off event. Note on events with velolicty=0 are
        handled like Note off events. The same mapping is applied for outgoing messages: `True` → Note on
        `False` → Note off. The velocity values for outgoing events can be configured via ``off_output_velocity`` and
        ``on_output_velocity``.

        The ``emulate_toggle`` parameter can be used to select a different operation mode, which is designed to turn a
        hardware controller's "flash" buttons (press sends Note on, release send note off) into toggle buttons. In this
        mode, an internal boolean state is maintained, which is toggled upon every incoming Note on events. Incoming
        Note off events are ignored. The state can be updated from within SHC by *writing* a bool value to the object.
        Outbound MIDI events are created, when the state is updated from within SHC (as in the normal operation mode)
        **and** as a response to incoming MIDI event, to make sure the hardware controller's button LEDs are always in
        the right state.

        The following table shows a comparison of the two operation modes:

        +------------------------------+--------------------------+-------------------------------------------+
        | incoming MIDI event sequence | ``emulate_toggle=False`` | ``emulate_toggle=True``                   |
        +==============================+==========================+===========================================+
        | *Note on*                    | ``True`` published       | ``True`` published, *Note on* sent back   |
        +------------------------------+--------------------------+-------------------------------------------+
        | *Note off*                   | ``False`` published      | ``True`` published, *Note on* sent back   |
        +------------------------------+--------------------------+-------------------------------------------+
        | *Note on*                    | ``True`` published       | ``False`` published, *Note off* sent back |
        +------------------------------+--------------------------+-------------------------------------------+
        | *Note off*                   | ``False`` published      | ``False`` published, *Note off* sent back |
        +------------------------------+--------------------------+-------------------------------------------+
        | …                            | …                        | …                                         |
        +------------------------------+--------------------------+-------------------------------------------+

        .. warning::

            For every note number, either :meth:`note_on_off` or :meth:`note_velocity` can be used, but not both.

        :param note: The MIDI note number to listen to/send events to [0 (C-2) - 127 (G9)].
        :param emulate_toggle: If True, emulate a toggle button (see above).
        :param off_output_velocity: The MIDI velocity of outbound *Note on* events. Defaults to 0.
        :param on_output_velocity: The MIDI velocity of outbound *Note on* events. Defaults to 127.
        :return: The *Connectible* object with the given specification
        """
        existing_variable = self._variable_map.get(('note_on', note))
        if existing_variable and isinstance(existing_variable, NoteOnOffVariable):
            return existing_variable
        elif existing_variable:
            raise ValueError("A variable for MIDI note {} is already existing but with a different variable type."
                             .format(note))
        variable = NoteOnOffVariable(self, note, emulate_toggle, off_output_velocity, on_output_velocity)
        self._variable_map[('note_on', note)] = variable
        self._variable_map[('note_off', note)] = variable
        return variable

    def control_change(self, control_channel: int) -> "ControlChangeVariable":
        """
        Create a *Connectable* object for this MIDI interface which handles *Control Change* events.

        The logics are pretty simple: The value of an incoming MIDI control change event for the given control number
        is published as a :class:`shc.datatypes.RangeUInt8` value. New values from SHC are sent out as control change
        MIDI events. The 7-bit MIDI values are converted to a RangeUInt8 by multiplying with 2 and adding 1 to all
        values >127. This way, the MIDI value 127 is mapped to ``RangeUInt8(255)``.

        :param control_channel: The MIDI control number (0-127)
        :return: The *Connectable* object with `RangeUInt8` type to interact with the specified MIDI control channel
        """
        existing_variable = self._variable_map.get(('control_change', control_channel))
        if existing_variable:
            assert isinstance(existing_variable, ControlChangeVariable)
            return existing_variable
        variable = ControlChangeVariable(self, control_channel=control_channel)
        self._variable_map[('control_change', control_channel)] = variable
        return variable

    def note_velocity(self, note: int) -> "NoteVelocityVariable":
        """
        Create a *Connectable* object for this MIDI interface which handles Note on/off events as range values using
        their velocity value.

        For each incoming Note on/off event for the specified note number the velocity value is published as a
        :class:`shc.datatypes.RangeUInt8`. The 7-bit MIDI values are converted to a RangeUInt8 by multiplying with 2 and
        adding 1 to all values >127. This way, the MIDI value 127 is mapped to ``RangeUInt8(255)``. When a new value
        is received from SHC (i.e. `written` to the object), an outbound Note on/off event is created, with a note
        velocity corresponding to the value. If the value is 0, a Note off event is created, otherwise a Note on event.

        .. warning::

            For every note number, either :meth:`note_on_off` or :meth:`note_velocity` can be used, but not both.

        :param note: The MIDI control number (0-127)
        :return: The *Connectable* object with `RangeUInt8` type to interact with the specified MIDI note.
        """
        existing_variable = self._variable_map.get(('note_on', note))
        if existing_variable and isinstance(existing_variable, NoteVelocityVariable):
            return existing_variable
        elif existing_variable:
            raise ValueError("A variable for MIDI note {} is already existing but with a different variable type."
                             .format(note))
        variable = NoteVelocityVariable(self, note=note)
        self._variable_map[('note_on', note)] = variable
        self._variable_map[('note_off', note)] = variable
        return variable

    def __repr__(self) -> str:
        return "{}(input_port_name={}, output_port_name={}, send_channel={}, receive_channel={})"\
            .format(self.__class__.__name__, self.input_port_name, self.output_port_name, self.send_channel,
                    self.receive_channel)


class AbstractMidiVariable(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _incoming_message(self, message: mido.Message) -> None:
        pass


class NoteVelocityVariable(Subscribable[RangeUInt8], Writable[RangeUInt8], AbstractMidiVariable):
    type = RangeUInt8

    def __init__(self, interface: MidiInterface, note: int):
        super().__init__()
        self.interface = interface
        self.note = note

    async def _write(self, value: RangeUInt8, origin: List[Any]) -> None:
        midi_value = value//2
        self.interface._send_message(mido.Message('note_off' if midi_value == 0 else 'note_on',
                                                  channel=self.interface.send_channel, note=self.note,
                                                  velocity=midi_value))

    def _incoming_message(self, message: mido.Message) -> None:
        value = message.velocity
        value = value * 2 + (1 if value > 63 else 0)
        self._publish(RangeUInt8(value), [])


class ControlChangeVariable(Subscribable[RangeUInt8], Writable[RangeUInt8], AbstractMidiVariable):
    type = RangeUInt8

    def __init__(self, interface: MidiInterface, control_channel: int):
        super().__init__()
        self.interface = interface
        self.control_channel = control_channel

    async def _write(self, value: RangeUInt8, origin: List[Any]) -> None:
        self.interface._send_message(mido.Message('control_change', channel=self.interface.send_channel,
                                                  control=self.control_channel, value=value//2))

    def _incoming_message(self, message: mido.Message) -> None:
        value = message.value
        value = value * 2 + (1 if value > 63 else 0)
        self._publish(RangeUInt8(value), [])


class NoteOnOffVariable(Subscribable[bool], Writable[bool], AbstractMidiVariable):
    type = bool

    def __init__(self, interface: MidiInterface, note: int, emulate_toggle: bool,
                 off_output_velocity: int, on_output_velocity: int):
        super().__init__()
        self.interface = interface
        self.note = note
        self.emulate_toggle = emulate_toggle
        self.off_velocity = off_output_velocity
        self.on_velocity = on_output_velocity

        # Required for emulate_toggle
        self.value = False

    async def _write(self, value: bool, origin: List[Any]) -> None:
        self.value = value
        self._to_midi(value)

    def _to_midi(self, value) -> None:
        self.interface._send_message(mido.Message('note_on' if value else 'note_off',
                                                  channel=self.interface.send_channel, note=self.note,
                                                  velocity=self.on_velocity if value else self.off_velocity))

    def _incoming_message(self, message: mido.Message) -> None:
        on = message.type == 'note_on' and message.velocity > 0
        if self.emulate_toggle:
            if on:
                self.value = not self.value
            self._to_midi(self.value)
        else:
            self.value = on
        self._publish(self.value, [])



import asyncio
import unittest

import shc


# Helper coroutine to trigger shutdown after one second automatically
async def shutdown() -> None:
    await asyncio.sleep(1)
    await shc.supervisor.stop()


class BasicTest(unittest.TestCase):
    def test_ui_showcase(self) -> None:
        shc.supervisor.event_loop.create_task(shutdown())
        import example.ui_showcase  # type: ignore
        shc.main()

    def test_ui_logging_showcase(self) -> None:
        shc.supervisor.event_loop.create_task(shutdown())

        import example.ui_logging_showcase  # type: ignore
        shc.main()

    def test_server_client_example(self) -> None:
        shc.supervisor.event_loop.create_task(shutdown())

        # The examples should actually be able to coexist in a single SHC instance
        import example.server_client.server  # type: ignore
        import example.server_client.client  # type: ignore
        shc.main()

    @classmethod
    def setUpClass(cls) -> None:
        shc.supervisor._REGISTERED_INTERFACES.clear()
        shc.timer.timer_supervisor.supervised_timers.clear()

    def tearDown(self) -> None:
        shc.supervisor._REGISTERED_INTERFACES.clear()
        shc.timer.timer_supervisor.supervised_timers.clear()



import datetime
import json
import unittest

from shc import conversion, Variable


class ConverterRegistrationTest(unittest.TestCase):
    def test_missing_converters(self) -> None:
        with self.assertRaises(TypeError):
            conversion.get_converter(float, datetime.date)
        with self.assertRaises(TypeError):
            json.dumps({'var': Variable(bool)}, cls=conversion.SHCJsonEncoder)
        with self.assertRaises(TypeError):
            fake_json_data = {'var': {'x': 'foobar'}}
            conversion.from_json(Variable, fake_json_data['var'])


class DefaultConverterTest(unittest.TestCase):
    def test_default_conversions(self) -> None:
        self.assertEqual(5, conversion.get_converter(float, int)(5.0))
        self.assertEqual(5.0, conversion.get_converter(float, int)(5))
        self.assertTrue(conversion.get_converter(float, bool)(5.0))
        self.assertTrue(conversion.get_converter(int, bool)(5))
        self.assertEqual("5.0", conversion.get_converter(float, str)(5.0))
        self.assertEqual("5", conversion.get_converter(int, str)(5))
        self.assertEqual(5, conversion.get_converter(str, int)("5"))
        self.assertEqual(5.3, conversion.get_converter(str, float)("5.3"))

    def test_default_json_converters(self) -> None:
        date1 = datetime.date(2020, 5, 17)
        datetime1 = datetime.datetime(2020, 5, 17, 17, 35, 19, 5005)
        datetime2 = datetime.datetime(2020, 5, 17, 17, 35, 19, 5005,
                                      tzinfo=datetime.timezone(datetime.timedelta(hours=2)))
        timedelta1 = datetime.timedelta(days=5, hours=7, minutes=3, milliseconds=25)

        serialized = json.dumps({
            'int': 5,
            'bool': True,
            'float': 5.3,
            'date1': date1,
            'datetime1': datetime1,
            'datetime2': datetime2,
            'timedelta1': timedelta1,
        }, cls=conversion.SHCJsonEncoder)

        deserialized = json.loads(serialized)

        self.assertEqual(5, conversion.from_json(int, deserialized['int']))
        self.assertEqual(True, conversion.from_json(bool, deserialized['bool']))
        self.assertEqual(5.3, conversion.from_json(float, deserialized['float']))
        self.assertEqual(date1, conversion.from_json(datetime.date, deserialized['date1']))
        self.assertEqual(datetime1, conversion.from_json(datetime.datetime, deserialized['datetime1']))
        self.assertEqual(datetime2, conversion.from_json(datetime.datetime, deserialized['datetime2']))
        self.assertEqual(timedelta1, conversion.from_json(datetime.timedelta, deserialized['timedelta1']))



import datetime
from typing import Optional, Type, Generic, List, Tuple

from ..base import T
from .generic import PersistenceVariable


class InMemoryPersistenceVariable(PersistenceVariable, Generic[T]):
    def __init__(self, type_: Type[T], keep: datetime.timedelta):
        super().__init__(type_, log=True)
        self.data: List[Tuple[datetime.datetime, T]] = []
        self.keep = keep

    async def _write_to_log(self, value: T):
        self.clean_up()
        self.data.append((datetime.datetime.now(datetime.timezone.utc), value))

    def clean_up(self) -> None:
        begin = datetime.datetime.now(datetime.timezone.utc) - self.keep
        keep_from: Optional[int] = None
        for i, (ts, _v) in enumerate(self.data):
            if ts > begin:
                keep_from = i
                break
        self.data = self.data[keep_from:]

    async def _read_from_log(self) -> Optional[T]:
        if not self.data:
            return None
        return self.data[-1][1]

    async def retrieve_log(self, start_time: datetime.datetime, end_time: datetime.datetime,
                           include_previous: bool = False) -> List[Tuple[datetime.datetime, T]]:
        iterator = iter(enumerate(self.data))
        try:
            start_index = next(i for i, (ts, _v) in iterator if ts >= start_time)
        except StopIteration:
            if include_previous and self.data:
                return self.data[-1:]
            else:
                return []
        if include_previous and start_index > 0:
            start_index -= 1
        try:
            end_index = next(i for i, (ts, _v) in iterator if ts > end_time)
        except StopIteration:
            return self.data[start_index:]
        return self.data[start_index:end_index]



import asyncio
import unittest
import unittest.mock

import shc.misc
from test._helper import ExampleSubscribable, ExampleWritable, async_test, ExampleReadable


class MiscTests(unittest.TestCase):

    @async_test
    async def test_two_way_pipe(self) -> None:
        pipe = shc.misc.TwoWayPipe(float)

        pub_left = ExampleSubscribable(float)
        pub_right = ExampleSubscribable(float)
        sub_left = ExampleWritable(float)
        sub_right = ExampleWritable(float)

        pipe.connect_left(pub_left)
        sub_left.connect(pipe)
        pipe.connect_right(pub_right)
        pipe.connect_right(sub_right)

        await pub_left.publish(42.0, [self])
        sub_right._write.assert_called_once_with(42.0, [self, pub_left, pipe.right])
        sub_left._write.assert_not_called()

        sub_right._write.reset_mock()
        await pub_right.publish(36.0, [self])
        sub_left._write.assert_called_once_with(36.0, [self, pub_right, pipe.left])
        sub_right._write.assert_not_called()

    @async_test
    async def test_two_way_pipe_concurrent_update(self) -> None:
        var1 = shc.Variable(int)
        pipe = shc.misc.TwoWayPipe(int).connect_left(var1)
        var2 = shc.Variable(int).connect(pipe.right)

        await asyncio.gather(var1.write(42, []), var2.write(56, []))
        self.assertEqual(await var1.read(), await var2.read())

    @async_test
    async def test_breakable_subscription_simple(self) -> None:
        pub = ExampleSubscribable(float)
        control = ExampleReadable(bool, True)
        sub = ExampleWritable(float)

        sub.connect(shc.misc.BreakableSubscription(pub, control))

        await pub.publish(42.0, [self])
        sub._write.assert_called_once_with(42.0, [self, pub, unittest.mock.ANY])

        sub._write.reset_mock()
        control.read.side_effect = (False,)
        await pub.publish(36.0, [self])
        sub._write.assert_not_called()

        sub._write.reset_mock()
        control.read.side_effect = (True,)
        await pub.publish(56.0, [self])
        sub._write.assert_called_once_with(56, unittest.mock.ANY)

    @async_test
    async def test_breakable_subscription_readsubscribable(self) -> None:
        pub = shc.Variable(float)
        control = shc.Variable(bool, initial_value=False)
        sub = ExampleWritable(float)

        sub.connect(shc.misc.BreakableSubscription(pub, control))

        # pub is uninitialized, so we should not receive anything, when control changes to True
        await control.write(True, [self])
        await asyncio.sleep(0.01)
        sub._write.assert_not_called()

        await pub.write(42.0, [self])
        await asyncio.sleep(0.01)
        sub._write.assert_called_once_with(42.0, [self, pub, unittest.mock.ANY])

        sub._write.reset_mock()
        await control.write(False, [self])
        await pub.write(56.0, [self])
        await asyncio.sleep(0.01)
        sub._write.assert_not_called()

        await control.write(True, [self])
        await asyncio.sleep(0.01)
        sub._write.assert_called_once_with(56.0, [self, control, unittest.mock.ANY])

    @async_test
    async def test_hysteresis(self) -> None:
        pub = ExampleSubscribable(float)
        hystersis = shc.misc.Hysteresis(pub, 42.0, 56.0)
        sub = ExampleWritable(bool).connect(hystersis)

        # Check initial value
        self.assertEqual(False, await hystersis.read())

        # Check climbing value
        await pub.publish(41.0, [self])
        await pub.publish(43.5, [self])
        await pub.publish(44.5, [self])
        self.assertEqual(False, await hystersis.read())
        sub._write.assert_not_called()

        await pub.publish(57.4, [self])
        sub._write.assert_called_once_with(True, [self, pub, hystersis])
        self.assertEqual(True, await hystersis.read())

        sub._write.reset_mock()
        await pub.publish(58, [self])
        sub._write.assert_not_called()
        self.assertEqual(True, await hystersis.read())

        # Check descending value
        await pub.publish(44.5, [self])
        self.assertEqual(True, await hystersis.read())
        sub._write.assert_not_called()

        await pub.publish(41.4, [self])
        sub._write.assert_called_once_with(False, [self, pub, hystersis])
        self.assertEqual(False, await hystersis.read())

        sub._write.reset_mock()
        await pub.publish(40.0, [self])
        sub._write.assert_not_called()
        self.assertEqual(False, await hystersis.read())

        # Check jumps
        await pub.publish(57.4, [self])
        sub._write.assert_called_once_with(True, [self, pub, hystersis])
        self.assertEqual(True, await hystersis.read())
        sub._write.reset_mock()
        await pub.publish(41.4, [self])
        sub._write.assert_called_once_with(False, [self, pub, hystersis])
        self.assertEqual(False, await hystersis.read())



#!/usr/bin/env python
import os.path
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='smarthomeconnect',
    version='0.3.1',
    description='The Smart Home Connect home automation framework based on AsyncIO',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Michael Thies',
    author_email='mail@mhthies.de',
    url='https://github.com/mhthies/smarthomeconnect',
    packages=find_packages(),
    include_package_data=True,
    python_requires='~=3.7',
    install_requires=[
        'aiohttp>=3.6,<4',
        'jinja2>=2.11,<3',
        'MarkupSafe>=1.1,<2',
    ],
    extras_require={
        'mysql': ['aiomysql==0.0.21'],
        'knx': ['knxdclient~=0.3.0'],
        'dmx': ['pyserial-asyncio==0.3'],
        'midi': ['mido>=1.2.9,<2', 'python-rtmidi>=1.4.6,<2'],
        'mqtt': ['paho-mqtt>=1.5.1,<2', 'asyncio-mqtt>=0.8.0,<0.9'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Home Automation',
        'Framework :: AsyncIO',
    ],
)



import asyncio
import json
import shutil
import subprocess
import time
import unittest
import unittest.mock
from contextlib import suppress
from typing import Dict, Any

import asyncio_mqtt  # type: ignore

import shc.interfaces.mqtt
import shc.interfaces.tasmota
from shc.datatypes import RGBWUInt8, RGBUInt8, RangeUInt8, RangeInt0To100
from test._helper import InterfaceThreadRunner, ExampleWritable, async_test


@unittest.skipIf(shutil.which("mosquitto") is None, "mosquitto MQTT broker is not available in PATH")
class TasmotaInterfaceTest(unittest.TestCase):
    def setUp(self) -> None:
        async def _construct(*args):
            return shc.interfaces.tasmota.TasmotaInterface(*args)

        self.client_runner = InterfaceThreadRunner(shc.interfaces.mqtt.MQTTClientInterface, "localhost", 42883)
        self.client = self.client_runner.interface
        self.interface: shc.interfaces.tasmota.TasmotaInterface = \
            asyncio.run_coroutine_threadsafe(_construct(self.client, 'test-device'), loop=self.client_runner.loop)\
            .result()
        self.broker_process = subprocess.Popen(["mosquitto", "-p", "42883"])
        time.sleep(0.25)

    def tearDown(self) -> None:
        asyncio.run_coroutine_threadsafe(self.interface.stop(), loop=self.client_runner.loop).result()
        self.client_runner.stop()
        self.broker_process.terminate()
        self.broker_process.wait()

    @async_test
    async def test_offline_state(self) -> None:
        task = asyncio.create_task(tasmota_device_mock("test-device"))
        await asyncio.sleep(0.25)

        try:
            target_offline = ExampleWritable(bool).connect(self.interface.online())
            self.client_runner.start()
            asyncio.run_coroutine_threadsafe(self.interface.start(), loop=self.client_runner.loop).result()
            await asyncio.sleep(0.25)

            target_offline._write.assert_called_once_with(True, unittest.mock.ANY)

            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

            await asyncio.sleep(0.25)
            target_offline._write.assert_called_with(False, unittest.mock.ANY)

        except Exception:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task
            raise

    @async_test
    async def test_color_external(self) -> None:
        def construct_color(r, g, b, w):
            return RGBWUInt8(RGBUInt8(RangeUInt8(r), RangeUInt8(g), RangeUInt8(b)), RangeUInt8(w))

        task = asyncio.create_task(tasmota_device_mock("test-device"))
        conn_color = self.interface.color_rgbw()
        target_color = ExampleWritable(RGBWUInt8).connect(conn_color)
        conn_power = self.interface.power()
        target_power = ExampleWritable(bool).connect(conn_power)
        await asyncio.sleep(0.25)

        try:
            self.client_runner.start()
            asyncio.run_coroutine_threadsafe(self.interface.start(), loop=self.client_runner.loop).result()

            await asyncio.sleep(0.25)
            target_color._write.assert_called_once_with(construct_color(0, 0, 0, 0), unittest.mock.ANY)
            target_power._write.assert_called_once_with(False, unittest.mock.ANY)

            async with asyncio_mqtt.Client("localhost", 42883, client_id="some-other-client") as c:
                await c.publish(f"cmnd/test-device/color", b'#aabbcc', retain=True)

            await asyncio.sleep(0.1)
            target_color._write.assert_called_with(construct_color(170, 187, 204, 0), [conn_color])  # rounding error
            target_power._write.assert_called_with(True, [conn_power])

        finally:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

    @async_test
    async def test_color_internal(self) -> None:
        def construct_color(r, g, b, w):
            return RGBWUInt8(RGBUInt8(RangeUInt8(r), RangeUInt8(g), RangeUInt8(b)), RangeUInt8(w))

        task = asyncio.create_task(tasmota_device_mock("test-device"))
        conn_color = self.interface.color_rgbw()
        target_color = ExampleWritable(RGBWUInt8).connect(conn_color)
        conn_power = self.interface.power()
        target_power = ExampleWritable(bool).connect(conn_power)
        conn_dimmer = self.interface.dimmer()
        target_dimmer = ExampleWritable(RangeInt0To100).connect(conn_dimmer)
        await asyncio.sleep(0.25)

        try:
            self.client_runner.start()
            asyncio.run_coroutine_threadsafe(self.interface.start(), loop=self.client_runner.loop).result()

            await asyncio.sleep(0.25)
            target_color._write.assert_called_once_with(construct_color(0, 0, 0, 0), unittest.mock.ANY)
            target_power._write.assert_called_once_with(False, unittest.mock.ANY)

            future = asyncio.wrap_future(asyncio.run_coroutine_threadsafe(
                conn_color.write(construct_color(175, 117, 48, 66), [self]), loop=self.client_runner.loop))
            # Should only return when the message returns from the broker
            await asyncio.sleep(0.001)
            self.assertFalse(future.done())
            await future

            target_color._write.assert_called_with(construct_color(175, 117, 48, 66), [self, conn_color])
            target_power._write.assert_called_with(True, [self, conn_power])
            target_dimmer._write.assert_called_with(RangeInt0To100(68), [self, conn_dimmer])

        finally:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

    @async_test
    async def test_sensor_ir(self) -> None:
        conn_ir = self.interface.ir_receiver()
        target_ir = ExampleWritable(bytes).connect(conn_ir)

        self.client_runner.start()
        asyncio.run_coroutine_threadsafe(self.interface.start(), loop=self.client_runner.loop).result()
        await asyncio.sleep(0.25)

        target_ir._write.assert_not_called()

        async with asyncio_mqtt.Client("localhost", 42883, client_id="some-other-client") as c:
            await c.publish(f"tele/test-device/RESULT",
                            json.dumps({"Time": "1970-01-10T03:12:43",
                                        "IrReceived": {"Protocol": "NEC", "Bits": 32, "Data": "0x00F7609F"}})
                            .encode('ascii'))

        await asyncio.sleep(0.25)
        target_ir._write.assert_called_once_with(b'\x00\xF7\x60\x9F', [conn_ir])


BASE_STATUS11: Dict[str, Dict[str, Any]] = {
    "StatusSTS":
        {"Time": "1970-01-08T04:46:42", "Uptime": "7T04:46:38", "UptimeSec": 621998, "Heap": 26,
         "SleepMode": "Dynamic", "Sleep": 50, "LoadAvg": 19, "MqttCount": 1, "POWER": "OFF", "Dimmer": 100,
         "Color": "00CCFFEE", "HSBColor": "192,100,100", "White": 93, "Channel": [0, 80, 100, 93],
         "Scheme": 0, "Fade": "OFF", "Speed": 1, "LedTable": "ON",
         "Wifi": {
             "AP": 1, "SSId": "some-network", "BSSId": "00:11:22:33:44:55", "Channel": 11, "RSSI": 76,
             "Signal": -62, "LinkCount": 1, "Downtime": "0T00:00:03"}}}


async def tasmota_device_mock(deviceid: str) -> None:
    # state variables:
    channel = [0, 0, 0, 0]
    power = False

    async with asyncio_mqtt.Client("localhost", 42883, client_id="FakeTasmotaDevice") as c:
        await c.subscribe(f'cmnd/{deviceid}/+')
        await c.publish(f"tele/{deviceid}/LWT", b'Online', retain=True)

        try:
            async with c.unfiltered_messages() as messages:
                async for msg in messages:
                    if msg.topic == f'cmnd/{deviceid}/status':
                        if msg.payload.strip() == b'11':
                            status = BASE_STATUS11.copy()
                            status['StatusSTS']['POWER'] = 'ON' if power else 'OFF'
                            status['StatusSTS']['Dimmer'] = int(max(channel)/255*100)
                            status['StatusSTS']['Channel'] = [int(v/255*100) for v in channel]
                            status['StatusSTS']['Color'] = \
                                '{:0>2X}{:0>2X}{:0>2X}{:0>2X}'.format(*channel)
                            status['StatusSTS']['White'] = int(channel[3]/255*100)
                            # TODO insert current HSBColor
                            await c.publish(f"stat/{deviceid}/STATUS11", json.dumps(status).encode('ascii'))
                        else:
                            pass
                            # not implemented

                    elif msg.topic == f'cmnd/{deviceid}/power':
                        power = msg.payload.lower() not in (b'0', 'off', 'false')
                        await c.publish(f'stat/{deviceid}/RESULT', b'{"POWER":"' + (b'ON' if power else b'OFF') + b'"}')
                        await c.publish(f'stat/{deviceid}/POWER', b'ON' if power else b'OFF')

                    elif msg.topic == f'cmnd/{deviceid}/color':
                        if msg.payload[0] == ord('#'):
                            data = bytes.fromhex(msg.payload.decode('ascii')[1:])
                            data += bytes([0] * (4 - len(data)))
                            channel = list(data[0:4])
                        else:
                            # not implemented
                            continue
                        power = max(channel) != 0
                        await c.publish(f'stat/{deviceid}/RESULT',
                                        json.dumps({"POWER": 'ON' if power else 'OFF',
                                                    "Dimmer": int(max(channel)/255*100),
                                                    "Color": '{:0>2X}{:0>2X}{:0>2X}{:0>2X}'
                                                             .format(*channel),
                                                    "HSBColor": "48,100,100",  # TODO
                                                    "White": int(channel[3]/255*100),
                                                    "Channel": [int(v/255*100) for v in channel]}).encode('ascii'))

        except asyncio.CancelledError:
            await c.publish(f"tele/{deviceid}/LWT", b'Offline', retain=True)
            raise



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""
This module provides SHC interface implementations to control DMX lighting equipment from SHC.

Simple usage example::

    import shc
    import shc.interfaces.dmx
    from shc.datatypes import RangeUInt8, RGBFloat1

    dmx_interface = shc.interfaces.dmx.EnttecDMXUSBProConnector(serial_url="/dev/ttyUSB3",
                                                                universe_size=10)  # improves transmission speed

    dimmer_1 = shc.Variable(RangeUInt8, "dimmer 1").connect(dmx_interface.address(1))
    dimmer_2 = shc.Variable(RangeUInt8, "dimmer 2").connect(dmx_interface.address(2))
    # ...

    rgb_par = shc.Variable(RGBUInt8, "LED PAR")
    rgb_par.red.connect(dmx_interface.address(7))
    rgb_par.green.connect(dmx_interface.address(8))
    rgb_par.blue.connect(dmx_interface.address(9))

(see :ref:`variables.tuple_field_access` for an explanation why accessing the RGB colors works this way.)
"""

import abc
import enum
import logging
from typing import List, Any, NamedTuple, Optional

import serial_asyncio  # type: ignore
import asyncio

from ..base import Writable
from ..datatypes import RangeUInt8
from ..supervisor import AbstractInterface

logger = logging.getLogger(__name__)


class AbstractDMXConnector(AbstractInterface, metaclass=abc.ABCMeta):
    """
    Abstract base class for DMX output interfaces.

    Concrete implementations, included with SHC, for specific protocols are:

    * :class:`EnttecDMXUSBProConnector`

    Instances of these classes provide a method :meth:`address` to create a `writable` object for a given DMX channel.
    """
    def __init__(self, universe_size: int = 512):
        super().__init__()
        self.universe = [0] * universe_size

    def address(self, dmx_address: int) -> "DMXAddress":
        """
        Get a *Writable* object for a specific DMX channel

        The expected value type is :class:`shc.datatypes.RangeUInt8`.

        :param dmx_address: The DMX channel address from 1 to 512
        :return: A *Writable* object for that DMX channel
        """
        return DMXAddress(self, dmx_address)

    @abc.abstractmethod
    async def transmit(self) -> None:
        """
        Abstract coroutine to trigger the transmission of the cached DMX universe to the DMX interface.

        This method is automatically called by :meth:`DMXAddress.write`.

        Concrete implementations of this method should make sure to await the successful transmission of the current
        universe state to the hardware interface. This might include waiting for the flushing of the serial buffer.
        """
        pass


class DMXAddress(Writable[RangeUInt8]):
    """
    `Writable` object of type :class:`RangeUInt8`, representing a single DMX channel of a :class:`AbstractDMXConnector`.

    Writing a value to a `DMXAddress` object sets the corresponding channel in the cached DMX universe to that value and
    triggers transmission of the universe to the hardware DMX interface.
    """
    type = RangeUInt8

    def __init__(self, connector: AbstractDMXConnector, address: int) -> None:
        self.address = address  # DMX channel address from 1 to 512
        self.connector = connector

    async def _write(self, value: RangeUInt8, origin: List[Any]) -> None:
        logger.debug("New value for DMX address %s: %s", self.address, value)
        self.connector.universe[self.address - 1] = value
        await self.connector.transmit()


class EnttecDMXUSBProConnector(AbstractDMXConnector):
    """
    A DMX Interface for the Enttec DMX USB PRO and compatible devices (with the same serial protocol).
    """
    # Build according to this spec:
    # https://web.archive.org/web/20200822142042/https://dol2kh495zr52.cloudfront.net/pdf/misc/dmx_usb_pro_api_spec.pdf

    def __init__(self, serial_url: str, universe_size: int = 512) -> None:
        super().__init__(universe_size)
        self.serial_url = serial_url
        self.running_transmit: Optional[asyncio.Future] = None
        self.next_transmit: Optional[asyncio.Future] = None

    async def start(self):
        logger.info("Starting Enttec DMX USB Pro interface on serial port %s ...", self.serial_url)
        self._reader, self._writer = await serial_asyncio.open_serial_connection(url=self.serial_url)
        await self._transmit()

    async def stop(self):
        logger.info("Closing serial port %s ...", self.serial_url)
        self._writer.close()
        await self._writer.wait_closed()

    @staticmethod
    def _universe_to_enttec(universe: List[int]) -> "EnttecMessage":
        """ Helper method to serialize a DMX universe (as List[int]) into an EnttacMessage to be sent to the DMX
        interface """
        DMX_LIGHTNING_DATA_START_CODE = 0
        data = bytes([DMX_LIGHTNING_DATA_START_CODE] + universe + [0] * (24 - len(universe)))
        return EnttecMessage(EntTecMessageLabel.OUTPUT_ONLY_SEND_DMX_PACKET, data)

    async def transmit(self) -> None:
        # In case, there is no running _transmit call, create a new one
        if not self.running_transmit or self.running_transmit.done():
            logger.debug("Immediately transmitting DMX data to interface ...")
            self.running_transmit = asyncio.create_task(self._transmit())
            await self.running_transmit
            logger.debug("DMX transmit finished ...")

        # In case there is already a running _transmit call, create a Future for the next _transmit call, await the
        # completion of the current _transmit, create a new _transmit call, and finally update the Future with its
        # result.
        elif not self.next_transmit:
            logger.debug("DMX transmit is already running. Queuing next transmit ...")
            self.next_transmit = asyncio.Future()
            await self.running_transmit
            logger.debug("Starting queued DMX transmit ...")
            self.running_transmit = self.next_transmit
            self.next_transmit = None
            try:
                await self._transmit()
                self.running_transmit.set_result(None)
                logger.debug("Queued DMX transmit finished ...")
            except Exception as e:
                self.running_transmit.set_exception(e)
                raise

        # In case there is a running _transmit and a Future for the next, simply await that future.
        else:
            logger.debug("DMX transmit is already running. Queued next transmit already exists. Wating for queued "
                         "transmit to finish ...")
            await self.next_transmit

    async def _transmit(self) -> None:
        """
        Internal helper coroutine to perform the actual transmission of the DMX universe to the DMX interface.

        This method converts the current state of the DMX universe to an EnttecMessage, writes it to the serial buffer
        and awaits the flush of the serial buffer to the interface."""
        # TODO catch connection errors
        self._writer.write(self._universe_to_enttec(self.universe).encode())
        await self._writer.drain()

    def __repr__(self) -> str:
        return "{}(serial_url={})".format(self.__class__.__name__, self.serial_url)


class EntTecMessageLabel(enum.Enum):
    REPROGRAM_FIRMWARE = 1
    PROGRAM_FLASH_PAGE = 2
    GET_WIDGET_PARAMETERS = 3
    SET_WIDGET_PARAMETERS = 4
    RECEIVED_DMX_PACKET = 5
    OUTPUT_ONLY_SEND_DMX_PACKET = 6
    SEND_RDM_PACKET = 7
    RECEIVE_DMX_ON_CHANGE = 8
    RECEIVED_DMX_CHANGE_OF_STATE_PACKET = 9
    GET_WIDGET_SERIAL_NUMBER = 10
    SEND_RDM_DISCOVERY = 11


class EnttecMessage(NamedTuple):
    label: EntTecMessageLabel
    data: bytes

    def encode(self) -> bytes:
        length = len(self.data)
        return bytes((0x7e, self.label.value, length & 0xff, (length >> 8) & 0xff)) + self.data + bytes((0xe7,))



import asyncio
import concurrent.futures
import functools
import heapq
import threading
import time
import unittest.mock
import datetime
from typing import Callable, Any, Awaitable, TypeVar, Generic, Type, List, Union, Tuple, Set, Optional

from shc import base


# #############################################
# General helper classes for testing async code
# #############################################
from shc.supervisor import AbstractInterface


def async_test(f: Callable[..., Awaitable[Any]]) -> Callable[..., Any]:
    """
    Decorator to transform async unittest coroutines into normal test methods
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(f(*args, **kwargs))
    return wrapper


class AsyncMock(unittest.mock.MagicMock):
    """
    Mock class which mimics an async coroutine (which can be called and then awaited) and an async context manager
    (which can be used in an `async with` block).

    This class is a simple replacement for unittest.mock.AsyncMock, which is only available since Python 3.8. This
    class does not have the assert_awaited features of the official AsyncMock.

    The async calls are passed to the normal call/enter/exit methods of the super class to use its usual builtin
    evaluation/assertion functionality (e.g. :meth:`unittest.mock.NonCallableMock.assert_called_with`).
    """
    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)

    async def __aenter__(self, *args, **kwargs):
        return self.__enter__(*args, **kwargs)

    async def __aexit__(self, *args, **kwargs):
        return self.__exit__(*args, **kwargs)


class ClockMock:
    """
    A mock/patch for the wall clock.

    When used as context managers, objects of this class patch the :meth:`datetime.datetime.now`,
    :meth:`datetime.date.today` and :meth:`time.time` to return the mocked time instead of the real wall clock time, was
    well as the :func:`time.sleep` and :func:`asyncio.sleep` functions to enhance the mocked time instead of actually
    sleeping. Optionally, the patched sleep methods can actually sleep for a predefined time to allow concurrent things
    to happen in the mean time (which would normally happen during the sleep). This also allows to let the ClockMock
    synchronize multiple threads or AsyncIO coroutines, sleeping for different times in parallel.

    For the purpose of patching `datetime.date.today` and `datetime.datetime.now`, these builtin classes need to be
    replaced by normal Python classes. For this purpose, the classes :class:`NewDate` and :class:`NewDateTime` are used,
    which inherit from their original pendants. To avoid failing type checks in other tests, the replacement is reverted
    when exiting the patch context.
    """
    class NewDate(datetime.date):
        pass

    class NewDateTime(datetime.datetime):
        pass

    def __init__(self, start_time: datetime.datetime, overshoot: datetime.timedelta = datetime.timedelta(),
                 actual_sleep: float = 0.0):
        self.current_time = start_time
        self.overshoot = overshoot
        self.actual_sleep = actual_sleep
        self.original_sleep = time.sleep
        self.original_async_sleep = asyncio.sleep
        self.original_time = time.time
        self.original_date = datetime.date
        self.original_datetime = datetime.datetime
        # A Mutex to make the the `queue` of sleeping Threads/Coroutines thread-safe
        self.mutex = threading.RLock()
        # Priority queue (heapq) of waiting Threads/Coroutines ordered by their wakeup time
        self.queue: List[Tuple[datetime.datetime, int, Union[int, asyncio.Task]]] = []
        # A counter for sleep times to make all entries in the queue unique
        self.counter = 0
        # Set of counter values of invalidated entries in the `queue`. Those must be skipped when checking if a
        # Thread/Couroutine is the next to wake up
        self.invalid_queue: Set[int] = set()
        # A real-world timestamp until which all sleeping threads/tasks shall actually sleep (via original_sleep/
        # original_async_sleep). This is required to always grant the `actual_sleep` time to a second thread/task, even
        # if a third thread/task has been sleeping before, checks the queue at the wrong time and is now the first in
        # the queue.
        self.actually_sleep_until = 0.0

    def tidy_queue(self):
        while True:
            if self.queue[0][1] in self.invalid_queue:
                self.invalid_queue.remove(self.queue[0][1])
                heapq.heappop(self.queue)
            else:
                return

    def sleep(self, seconds: float) -> None:
        thread_id = threading.get_ident()
        # Add us to the queue of waiting threads/coroutines
        with self.mutex:
            target_time = self.current_time + datetime.timedelta(seconds=seconds) + self.overshoot
            heapq.heappush(self.queue, (target_time, self.counter, thread_id))
            self.counter += 1
            self.actually_sleep_until = self.original_time() + self.actual_sleep
            sleep_for = self.actual_sleep
        # Actually sleep while other thread/coroutine work or sleep (but shorter than we do in emulated time)
        while True:
            self.original_sleep(sleep_for)
            with self.mutex:
                self.tidy_queue()
                sleep_for = self.actually_sleep_until - self.original_time()
                if sleep_for <= 0 and self.queue[0][2] == thread_id:
                    heapq.heappop(self.queue)
                    break
        # Update emulated wall clock to the target sleep time
        self.current_time = target_time

    async def async_sleep(self, seconds: float) -> None:
        current_task = asyncio.current_task()
        target_time = self.current_time + datetime.timedelta(seconds=seconds) + self.overshoot
        # Add us to the queue of waiting threads/coroutines
        with self.mutex:
            count = self.counter
            heapq.heappush(self.queue, (target_time, count, current_task))  # type: ignore
            self.counter += 1
            self.actually_sleep_until = self.original_time() + self.actual_sleep
            sleep_for = self.actual_sleep
        # Actually sleep while other thread/coroutine work or sleep (but shorter than we do in emulated time)
        while True:
            try:
                await self.original_async_sleep(sleep_for)
            except asyncio.CancelledError:
                # If the sleep is ended by cancelling the task, we must make sure to invalidate our queue entry.
                # Otherwise, other threads/coroutines would sleep infinitely waiting for us to wake up
                with self.mutex:
                    self.invalid_queue.add(count)
                raise
            # Check if we are next to wake up or if another thread/coroutine should wake up before us, according to
            # emulated time
            with self.mutex:
                self.tidy_queue()
                sleep_for = self.actually_sleep_until - self.original_time()
                if sleep_for <= 0 and self.queue[0][2] is current_task:
                    heapq.heappop(self.queue)
                    break
        # Update emulated wall clock to the target sleep time
        self.current_time = target_time

    def now(self, tz=None) -> datetime.datetime:
        time_ = self.current_time
        if tz is not None:
            return time_.astimezone(tz)
        return time_

    def time(self) -> float:
        return self.current_time.timestamp()

    def today(self) -> datetime.date:
        return self.current_time.date()

    def __enter__(self) -> "ClockMock":
        import datetime
        datetime.date = self.NewDate  # type: ignore
        datetime.datetime = self.NewDateTime  # type: ignore

        self.patches = (
            unittest.mock.patch('time.sleep', new=self.sleep),
            unittest.mock.patch('asyncio.sleep', new=self.async_sleep),
            unittest.mock.patch('datetime.datetime.now', new=self.now),
            unittest.mock.patch('time.time', new=self.time),
            unittest.mock.patch('datetime.date.today', new=self.today),
        )
        for p in self.patches:
            p.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for p in self.patches:
            p.__exit__(exc_type, exc_val, exc_tb)

        datetime.datetime = self.original_datetime
        datetime.date = self.original_date


# ###########################
# Example Connectable objects
# ###########################

T = TypeVar('T')


class ExampleReadable(base.Readable[T], Generic[T]):
    read: AsyncMock  # required to let MyPy know that we can use Mock's methods

    def __init__(self, type_: Type[T], value: T, side_effect=None):
        self.type = type_
        super().__init__()
        self.read = AsyncMock(return_value=value, side_effect=side_effect)

    async def read(self) -> T: ...  # type: ignore


class ExampleSubscribable(base.Subscribable[T], Generic[T]):
    def __init__(self, type_: Type[T]):
        self.type = type_
        super().__init__()

    async def publish(self, val: T, origin: List[Any]) -> None:
        await self._publish_and_wait(val, origin)


class ExampleWritable(base.Writable[T], Generic[T]):
    _write: AsyncMock  # required to let MyPy know that we can use Mock's methods

    def __init__(self, type_: Type[T]):
        self.type = type_
        super().__init__()
        self._write = AsyncMock()

    async def _write(self, value: T, origin: List[Any]) -> None: ...  # type: ignore


class ExampleReading(base.Reading[T], Generic[T]):
    def __init__(self, type_: Type[T], optional: bool):
        self.is_reading_optional = optional
        self.type = type_
        super().__init__()

    async def do_read(self) -> Optional[T]:
        return await self._from_provider()


class SimpleIntRepublisher(base.Writable, base.Subscribable):
    type = int

    async def _write(self, value: T, origin: List[Any]):
        await self._publish_and_wait(value, origin)


IT = TypeVar('IT', bound=AbstractInterface)


class InterfaceThreadRunner(Generic[IT]):
    """
    Some magic for running an SHC interface in a separate AsyncIO event loop in a background thread. This is helpful
    for testing the interface's features from the main thread, using blocking functions (e.g. selenium for web testing).

    The interface will most probably not be thread-safe internally (as SHC usually runs in only one AsyncIO event loop
    in a single thread). Thus, its methods/coroutines must only be called within the `InterfaceThreadRunner`s AsnycIO
    event loop. For convenience, there is a :meth:`start` method that calls the interface's start method correctly::

        runner = InterfaceThreadRunner(SomeSHCInterfaceClass, "arg1", "arg2")
        runner.start()
        asyncio.run_coroutine_threadsafe(interface.some_coro(*args), loop=runner.loop).result()
        runner.loop.call_soon_threadsafe(interface.some_method(*args))
        runner.stop()

    :ivar loop: The event loop of the background thread, in which the interface is running. This variable is only
        available after :meth:`start` has successfully completed.
    :ivar interface: The constructed interface
    :param interface_class: The interface to create an instance of. An interface of this class is constructed in the
        context of the background thread (to assign the correct asyncio Event loop to any Futures, Events, etc.) and
        available in the `interface` attribute after construction.
    :param args: positional arguments to be passed to the interface's constructor
    :param kwargs: keyword arguments to be passed to the interface's constructor
    :raises TimeoutError: When the interface cannot be constructed within 5 seconds.
    """
    executor = concurrent.futures.ThreadPoolExecutor()

    def __init__(self, interface_class: Type[IT], *args, **kwargs):
        self._server_constructed_future: concurrent.futures.Future[None] = concurrent.futures.Future()
        self.started = False
        self.interface: IT
        self.future = self.executor.submit(self._run, interface_class, args, kwargs)
        self._server_constructed_future.result(timeout=5)

    def _run(self, interface_class: Type[IT], args, kwargs) -> None:
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        try:
            self.interface = interface_class(*args, **kwargs)  # type: ignore
            self._server_constructed_future.set_result(None)
        except Exception as e:
            self._server_constructed_future.set_exception(e)
            return
        self._stopped_event = asyncio.Event()
        self.started = True
        self.loop.run_until_complete(self._stopped_event.wait())
        self.loop.close()

    def start(self) -> None:
        start_future = asyncio.run_coroutine_threadsafe(self.interface.start(), self.loop)
        start_future.result(timeout=5)

    def stop(self) -> None:
        """
        Stop the interface via its stop() coroutine and block until it is fully shutdown.
        """
        if not self.started:
            return
        self.started = False
        stop_future = asyncio.run_coroutine_threadsafe(self.interface.stop(), self.loop)
        stop_future.result(timeout=5)
        self.loop.call_soon_threadsafe(self._stopped_event.set)
        self.future.result(timeout=5)
